/*
Copyright 2020-2023 CAcert Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package handlers

import (
	"html/template"
	"net/http"

	client "github.com/ory/hydra-client-go/v2"
	log "github.com/sirupsen/logrus"

	"code.cacert.org/cacert/oidc-idp/internal/services"
)

type LogoutHandler struct {
	adminClient client.OAuth2Api
	logger      *log.Logger
}

func (h *LogoutHandler) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	challenge := r.URL.Query().Get("logout_challenge")
	h.logger.WithField("challenge", challenge).Debug("received logout challenge")

	logoutRequest, response, err := h.adminClient.GetOAuth2LogoutRequest(
		r.Context(),
	).LogoutChallenge(challenge).Execute()
	if err != nil {
		h.logger.WithError(err).Error("error getting logout requests")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)

		return
	}

	defer func() { _ = response.Body.Close() }()

	h.logger.WithFields(
		log.Fields{"response": response.Status, "logout_request": logoutRequest},
	).Debug("got response for GetOAuth2LogoutRequest")

	acceptLogoutRequest, response, err := h.adminClient.AcceptOAuth2LogoutRequest(
		r.Context(),
	).LogoutChallenge(challenge).Execute()
	if err != nil {
		h.logger.WithError(err).Error("accept logout request failed")
		http.Error(w, http.StatusText(http.StatusInternalServerError), http.StatusInternalServerError)
	}

	defer func() { _ = response.Body.Close() }()

	h.logger.WithFields(
		log.Fields{"response": response.Status, "accept_logout_request": acceptLogoutRequest},
	).Debug("got response for AcceptOAuth2LogoutRequest")

	w.Header().Set("Location", acceptLogoutRequest.GetRedirectTo())
	w.WriteHeader(http.StatusFound)
}

func NewLogout(logger *log.Logger, adminClient client.OAuth2Api) *LogoutHandler {
	return &LogoutHandler{
		logger:      logger,
		adminClient: adminClient,
	}
}

type LogoutSuccessHandler struct {
	logger    *log.Logger
	trans     *services.I18NService
	templates TemplateCache
}

func (h *LogoutSuccessHandler) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	if r.Method != http.MethodGet {
		http.Error(w, http.StatusText(http.StatusMethodNotAllowed), http.StatusMethodNotAllowed)

		return
	}

	localizer := getLocalizer(h.trans, r)

	h.templates.render(h.logger, w, LogoutSuccessful, map[string]interface{}{
		"Title": h.trans.LookupMessage("LogoutSuccessfulTitle", nil, localizer),
		"Explanation": template.HTML(h.trans.LookupMarkdownMessage( //nolint:gosec
			"LogoutSuccessfulText",
			nil,
			localizer,
		)),
	})
}

func NewLogoutSuccess(
	logger *log.Logger,
	templateCache TemplateCache,
	trans *services.I18NService,
) *LogoutSuccessHandler {
	return &LogoutSuccessHandler{logger: logger, trans: trans, templates: templateCache}
}
