/*
Copyright 2020-2023 CAcert Inc.
SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package handlers

import (
	"crypto/x509"
	"fmt"
	"net/http"
	"time"

	log "github.com/sirupsen/logrus"
)

func EnableHSTS() func(http.Handler) http.Handler {
	return func(next http.Handler) http.Handler {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			const Days180 = 180
			w.Header().Set("Strict-Transport-Security", fmt.Sprintf("max-age=%d", int((time.Hour*24*Days180).Seconds())))
			next.ServeHTTP(w, r)
		})
	}
}

func getDataFromClientCert(logger *log.Logger, r *http.Request) (string, []string) {
	if r.TLS != nil && r.TLS.PeerCertificates != nil && len(r.TLS.PeerCertificates) > 0 {
		firstCert := r.TLS.PeerCertificates[0]

		if !isClientCertificate(firstCert) {
			return "", nil
		}

		for _, email := range firstCert.EmailAddresses {
			logger.WithField(
				"email", email,
			).Info("authenticated with a client certificate for email address")
		}

		return firstCert.Subject.CommonName, firstCert.EmailAddresses
	}

	return "", nil
}

func isClientCertificate(cert *x509.Certificate) bool {
	for _, ext := range cert.ExtKeyUsage {
		if ext == x509.ExtKeyUsageClientAuth {
			return true
		}
	}

	return false
}
